// XXX copytight = ?
// based on:
// https://groups.google.com/d/msg/golang-nuts/PYWxjT2v6ps/dL71oJk1mXEJ
// https://play.golang.org/p/f9HY6-z8Pp

package main
// weak reference

import (
	"runtime"
	"sync"
	"unsafe"
)

// iface is how Go runtime represents an interface.
//
// NOTE layout must be synchronized to Go runtime representation.
type iface struct {
	typ  uintptr // type
	data uintptr // data
}

type weakRefState int32

const (
	objGot      weakRefState = +1 // WeakRef.Get returned !nil
	objLive     weakRefState =  0 // object is alive, Get did not run yet in this GC cycle
	objReleased weakRefState = -1 // the finalizer marked object as released
)


// WeakRef is a weak reference.
//
// Create one with NewWeakRef and retrieve referenced object with Get.
type WeakRef struct {
	iface

	// XXX try to do without mutex and only with atomics
	mu    sync.Mutex
	state weakRefState
}

// XXX
func NewWeakRef(obj interface{}) *WeakRef {
	// original code uses ^ for pointers not to cause GC thinking that original
	// object is still referenced from WeakRef.data. However starting with ~ Go1.4
	// the GC is precise, so we don't need that trick this days.
	i := (*[2]uintptr)(unsafe.Pointer(&obj))
	w := &WeakRef{
		iface: *(*iface)(unsafe.Pointer(&obj)),
		state: objLive,
	}

	var release func(interface{})
	release = func(_ interface{}) {
		// GC decided that the object is no longer reachable and
		// scheduled us to run as finalizer. During the time till we
		// actually run, WeakRef.Get might have been come to run and
		// "rematerializing" the object for use. Check if we do not
		// race with any Get in progress, and reschedule us to retry at
		// next GC if we do.
		w.mu.Lock()
		if w.state == objGot {
			w.state = objLive
			runtime.SetFinalizer(obj, release)
		} else {
			w.state = objReleased
		}
		w.mu.Unlock()
	}

	runtime.SetFinalizer(obj, release)
	return w
}

// Get returns object pointed to by this weak reference.
//
// If original object is still alive - it is returned.
// If not - nil is returned.
func (w *WeakRef) Get() (obj interface{}) {
	w.mu.Lock()
	if w.state != objReleased {
		w.state = objGot

		// recreate interface{} from saved words.
		// XXX do writes as pointers so that compiler emits write barriers to notify GC?
		i := (*iface)(unsafe.Pointer(&obj))
		*i = w.iface
	}
	w.mu.Unlock()
	return obj
}
